__name__ = "versipy"
__version__ = "0.2.4.post1"
__description__ = "Versatile version and medatada managment across the python packaging ecosystem with git integration"
__url__ = "https://github.com/a-slide/versipy"
__licence__ = "GPLv3"
__author__ = "Adrien Leger"
